﻿using System;
using System.Collections.ObjectModel;
using System.Drawing;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using ThinkGeo.MapSuite.Core;

namespace CacheGenerator
{
    public partial class MainForm : Form
    {
        delegate void InvokeMethod();
        DateTime startTime;
        TileCacheGenerator tileCacheGenerator;
        GeographyUnit mapUnit = GeographyUnit.Meter;

        public MainForm()
        {
            InitializeComponent();
        }

        private void CachedImagesGenerateroForm_Load(object sender, EventArgs e)
        {
            cmbStartZoomLevel.Text = "ZoomLevel01";
            cmbEndZoomLevel.Text = "ZoomLevel07";
            cmbTileImageType.Text = "PNG";
        }

        private Collection<double> GetScalesToCache()
        {
            Collection<double> scales = LayerProvider.GetScalesToCache(mapUnit);

            if (cmbStartZoomLevel.SelectedIndex > cmbEndZoomLevel.SelectedIndex)
            {
                cmbStartZoomLevel.SelectedIndex = cmbEndZoomLevel.SelectedIndex;
            }
            Collection<double> scalesToCache = new Collection<double>();
            for (int i = cmbStartZoomLevel.SelectedIndex; i <= cmbEndZoomLevel.SelectedIndex; i++)
            {
                scalesToCache.Add(scales[i]);
            }

            return scalesToCache;
        }

        private void btnGenerate_Click(object sender, EventArgs e)
        {
            btnGenerate.Enabled = false;
            tileCacheGenerator = new TileCacheGenerator();
            tileCacheGenerator.ProgressChanged += new EventHandler<CacheGeneratorProgressChangedEventArgs>(tileCacheGenerator_ProgressChanged);
            tileCacheGenerator.GenerationCompleted += new EventHandler<EventArgs>(tileCacheGenerator_GenerationCompleted);

            if (cmbTileImageType.Text.ToUpperInvariant() == "PNG")
            {
                tileCacheGenerator.TileImageFormat = TileImageFormat.Png;
            }
            else if (cmbTileImageType.Text.ToUpperInvariant() == "JPEG")
            {
                tileCacheGenerator.TileImageFormat = TileImageFormat.Jpeg;
                tileCacheGenerator.JpegQuality = Int16.Parse(txtJpegQuality.Text);
            }
            string[] upperLeftPointInString = txtUpperLeftX.Text.Split(',');
            string[] lowerRightPointInString = txtLowerRightX.Text.Split(',');
            tileCacheGenerator.CachingExtent = new RectangleShape(double.Parse(upperLeftPointInString[0]), double.Parse(upperLeftPointInString[1]), double.Parse(lowerRightPointInString[0]), double.Parse(lowerRightPointInString[1]));

            if (checkBoxWatermark.Checked)
            {
                tileCacheGenerator.WatermarkBitmap = new Bitmap(txtWatermarkPath.Text);
            }

            tileCacheGenerator.MapUnit = mapUnit;
            tileCacheGenerator.CacheFolder = txtCacheFolder.Text;

            startTime = System.DateTime.Now;
            tileCacheGenerator.ScalesToCache = GetScalesToCache();
            tileCacheGenerator.LayersToCache = LayerProvider.GetLayersToCache();

            new Thread(() =>
            {
                tileCacheGenerator.GenerateTiles();
            }) { IsBackground = true }.Start();
        }

        void tileCacheGenerator_ProgressChanged(object sender, CacheGeneratorProgressChangedEventArgs e)
        {
            // Make the label update every 100 tiles. 
            if (e.CurrentTileIndex % 100 != 0 && e.CurrentTileIndex < e.TotalTilesCount)
            {
                return;
            }
            TimeSpan timeSpan = DateTime.Now - startTime;

            lblCurrentImageCount.Invoke((InvokeMethod)delegate()
            {
                lblCurrentImageCount.Text = string.Format("Current Progress: {0} %", e.ProgressPercentage);
            });

            lblTileStatus.Invoke((InvokeMethod)delegate
            {
                lblTileStatus.Text = String.Format("{0:N0} Tiles Generated. ", e.CurrentTileIndex);
            });

            double totalMilliseconds = timeSpan.TotalMilliseconds;
            double averageMilliseconds = totalMilliseconds / e.CurrentTileIndex;
            double remainingTime = (e.TotalTilesCount - e.CurrentTileIndex) * averageMilliseconds;
            if (remainingTime < 0)
            {
                remainingTime = 0;
            }

            lblTotalTime.Invoke((InvokeMethod)delegate()
            {
                lblTotalTime.Text = String.Format("TotalTime: {0}; RemainingTime: {1}; AverageTime: {2} ms/Tile", FormattedTimeString(timeSpan), FormattedTimeString(TimeSpan.FromMilliseconds(remainingTime)), averageMilliseconds.ToString("N1"));
            });
        }

        void tileCacheGenerator_GenerationCompleted(object sender, EventArgs e)
        {
            btnGenerate.Invoke((InvokeMethod)delegate()
            {
                btnGenerate.Enabled = true;
            });

            lblCurrentImageCount.Invoke((InvokeMethod)delegate()
            {
                lblCurrentImageCount.Text = "Current Progress: 100%";
            });

            MessageBox.Show("Generation Completed!");
        }

        public static string FormattedTimeString(TimeSpan ts)
        {
            if (ts.Days > 0)
            {
                return String.Format("{0} Days {1} Hours", ts.Days.ToString(), ts.Hours.ToString());
            }
            else if (ts.Hours > 0)
            {
                return String.Format("{0} Hours {1} Minutes", ts.Hours.ToString(), ts.Minutes.ToString());
            }
            else if (ts.Minutes > 0)
            {
                return String.Format("{0} Minutes {1} Seconds", ts.Minutes.ToString(), ts.Seconds.ToString());
            }
            else
            {
                return String.Format("{0} Seconds", ts.Seconds.ToString());
            }
        }

        private void btnPreview_Click(object sender, EventArgs e)
        {
            TileCacheGenerator tileCacheGenerator = new TileCacheGenerator();
            tileCacheGenerator.MapUnit = mapUnit;
            tileCacheGenerator.ScalesToCache = GetScalesToCache();

            string[] upperLeftPointInString = txtUpperLeftX.Text.Split(',');
            string[] lowerRightPointInString = txtLowerRightX.Text.Split(',');
            tileCacheGenerator.CachingExtent = new RectangleShape(double.Parse(upperLeftPointInString[0]), double.Parse(upperLeftPointInString[1]), double.Parse(lowerRightPointInString[0]), double.Parse(lowerRightPointInString[1]));

            Collection<long> tilesCount = tileCacheGenerator.GetTilesCountsForScales();
            long totalTilesCount = 0;

            int startIndex = cmbStartZoomLevel.SelectedIndex;
            StringBuilder sb = new StringBuilder();
            sb.AppendLine();
            sb.AppendLine();

            for (int i = 0; i < tilesCount.Count; i++)
            {
                totalTilesCount += tilesCount[i];
                sb.AppendFormat("Tiles Count In {0}: {1:N0}", cmbStartZoomLevel.Items[startIndex++], tilesCount[i]);
                sb.AppendLine();
            }

            sb.Insert(0, string.Format("Total Tiles Count: {0:N0}", totalTilesCount));

            MessageBox.Show(sb.ToString());
        }

        private void btnSelectCacheFolder_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog folderBrowserDialog = new FolderBrowserDialog();
            if (folderBrowserDialog.ShowDialog() == DialogResult.OK)
            {
                txtCacheFolder.Text = folderBrowserDialog.SelectedPath;
            }
        }

        private void btnSelectWatermark_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.InitialDirectory = Application.StartupPath;
            openFileDialog.Filter = "Image File(*.jpg,*.gif,*.bmp,*.png)|*.jpg;*.gif;*.bmp;*.png|All File(*.*)|*.*";
            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                txtWatermarkPath.Text = openFileDialog.FileName;
            }
        }

        private void cmbTileImageType_SelectedIndexChanged(object sender, EventArgs e)
        {
            txtJpegQuality.Enabled = (cmbTileImageType.Text == "JPEG");
            lblJpegQuality.Enabled = txtJpegQuality.Enabled;
        }

        private void btnCancel_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void checkBoxWatermark_CheckedChanged(object sender, EventArgs e)
        {
            txtWatermarkPath.ReadOnly = !checkBoxWatermark.Checked;
        }
    }
}
