﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Drawing;
using System.Linq;
using System.Threading.Tasks;
using ThinkGeo.MapSuite.Core;

namespace CacheGenerator
{
    class TileCacheGenerator
    {
        private int tileWidth = 256;
        private int tileHeight = 256;
        private short jpegQuality;
        private TileImageFormat tileImageFormat;
        private Collection<Layer> layersToCache;
        private GeographyUnit mapUnit;
        private Collection<double> scales;
        private RectangleShape cachingExtent;
        private string cacheFolder;
        private Bitmap watermarkBitmap;
        private int currentTileIndex = 0;
        private long totalTilesCount = 0;

        public event EventHandler<CacheGeneratorProgressChangedEventArgs> ProgressChanged;
        public event EventHandler<EventArgs> GenerationCompleted;
        
        static object locker = new object();

        public int TileWidth
        {
            get { return tileWidth; }
            set { tileWidth = value; }
        }

        public int TileHeight
        {
            get { return tileHeight; }
            set { tileHeight = value; }
        }

        public TileImageFormat TileImageFormat
        {
            get { return tileImageFormat; }
            set { tileImageFormat = value; }
        }

        public short JpegQuality
        {
            get { return jpegQuality; }
            set { jpegQuality = value; }
        }

        public Collection<Layer> LayersToCache
        {
            get { return layersToCache; }
            set { layersToCache = value; }
        }

        public GeographyUnit MapUnit
        {
            get { return mapUnit; }
            set { mapUnit = value; }
        }

        public Collection<double> ScalesToCache
        {
            get { return scales; }
            set { scales = value; }
        }

        public RectangleShape CachingExtent
        {
            get { return cachingExtent; }
            set { cachingExtent = value; }
        }

        public string CacheFolder
        {
            get { return cacheFolder; }
            set { cacheFolder = value; }
        }

        public Bitmap WatermarkBitmap
        {
            get { return watermarkBitmap; }
            set { watermarkBitmap = value; }
        }

        public void GenerateTiles()
        {
            totalTilesCount = GetTilesCount();

            Parallel.ForEach(scales, (scale) =>
            {
                GenerateTilesForOneScale(scale);
            });

            foreach (Layer layer in layersToCache)
            {
                layer.Close();
            }

            OnGenerationCompleted(new EventArgs());
        }

        private void GenerateTilesForOneScale(double scale)
        {
            FileBitmapTileCache tileCache = new FileBitmapTileCache(cacheFolder, string.Empty);
            tileCache.ImageFormat = this.tileImageFormat;
            tileCache.JpegQuality = this.jpegQuality;
            tileCache.TileMatrix = new MapSuiteTileMatrix((double)scale, tileWidth, tileHeight, mapUnit);

            Collection<TileMatrixCell> cells = tileCache.TileMatrix.GetIntersectingCells(cachingExtent);

            Parallel.ForEach(cells, (cell) =>
            {
                Bitmap bitmap = new Bitmap(tileWidth, tileHeight);
                DrawOnBitmap(layersToCache, mapUnit, cell.BoundingBox, bitmap);
                tileCache.SaveTiles(bitmap, cell.BoundingBox);
                lock (locker)
                {
                    currentTileIndex++;
                    int percentage = (int)(currentTileIndex * 100 / totalTilesCount);
                    if (percentage > 100)
                    {
                        percentage = 100;
                    }
                    OnProgressChanged(new CacheGeneratorProgressChangedEventArgs(percentage, null, currentTileIndex, totalTilesCount));
                }
            });
        }

        private void DrawOnBitmap(IEnumerable<Layer> layersTmp, GeographyUnit mapUnit, RectangleShape extent, Bitmap bitmap)
        {
            GeoCanvas canvas = new GdiPlusGeoCanvas();
            canvas.BeginDrawing(bitmap, extent, mapUnit);
            Collection<Layer> layers = LayerProvider.GetLayersToCache();
            foreach (Layer layer in layers)
            {
                layer.Open();
                layer.Draw(canvas, new Collection<SimpleCandidate>());
            }
            Bitmap watermark = GetWatermarkBitmap();
            if (watermark != null)
            {
                Graphics g = Graphics.FromImage(bitmap);
                g.DrawImage(watermark, 0, 0, bitmap.Width, bitmap.Height);
                g.Dispose();
            }
            canvas.EndDrawing();
        }

        public long GetTilesCount()
        {
            Collection<long> cells = GetTilesCountsForScales();
            return cells.Sum();
        }

        public Collection<long> GetTilesCountsForScales()
        {
            Collection<long> cells = new Collection<long>();
            foreach (double scale in scales)
            {
                MapSuiteTileMatrix tileMatrix = new MapSuiteTileMatrix(scale, tileWidth, tileHeight, mapUnit);
                RowColumnRange rowColumnRange = tileMatrix.GetIntersectingRowColumnRange(cachingExtent);
                long deltaColumn = rowColumnRange.MaxColumnIndex - rowColumnRange.MinColumnIndex + 1;
                long deltaRow = rowColumnRange.MaxRowIndex - rowColumnRange.MinRowIndex + 1;

                long cellsCountUnderScale = deltaColumn * deltaRow;

                cells.Add(cellsCountUnderScale);
            }
            return cells;
        }

        protected virtual void OnProgressChanged(CacheGeneratorProgressChangedEventArgs e)
        {
            EventHandler<CacheGeneratorProgressChangedEventArgs> handler = ProgressChanged;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        protected virtual void OnGenerationCompleted(EventArgs e)
        {
            EventHandler<EventArgs> handler = GenerationCompleted;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        private Bitmap GetWatermarkBitmap()
        {
            if (watermarkBitmap == null)
            {
                return null;
            }

            Bitmap clonedWatermarkBitmap = null;
            lock (watermarkBitmap)
            {
                clonedWatermarkBitmap = (Bitmap)watermarkBitmap.Clone();
            }
            return clonedWatermarkBitmap;
        }
    }
}